# -*- coding: utf-8 -*-


# Built-in
import warnings


# Common
import numpy as np
from matplotlib.tri import Triangulation as mplTri


# specific
from . import _generic_check


_ELEMENTS = 'knots'


# #############################################################################
# #############################################################################
#                           mesh generic check
# #############################################################################


def _mesh2D_check(
    coll=None,
    domain=None,
    res=None,
    R=None,
    Z=None,
    knots=None,
    cents=None,
    trifind=None,
    key=None,
):

    # key
    key = _generic_check._check_var(
        key, 'key',
        types=str,
        excluded=list(coll.dobj.get('mesh', {}).keys())
    )

    # rect of tri ?
    lc = [
        domain is not None or (R is not None and Z is not None),
        knots is not None and cents is not None,
    ]
    if all(lc) or not any(lc):
        msg = (
            "Either domain xor (R, Z) xor (knots, cents) must be provided!\n"
            "Provided:\n"
            f"\t- domain, res: {domain}, {res}\n"
            f"\t- type(R), type(Z): {type(R)}, {type(Z)}\n"
            f"\t- type(knots), type(cents): {type(knots)}, {type(cents)}\n"
        )
        raise Exception(msg)

    elif lc[0]:
        dref, dmesh = _mesh2DRect_to_dict(
            domain=domain,
            res=res,
            R=R,
            Z=Z,
            key=key,
        )
        ddata = None

    elif lc[1]:
        dref, ddata, dmesh = _mesh2DTri_to_dict(
            knots=knots,
            cents=cents,
            trifind=trifind,
            key=key,
        )
    return dref, ddata, dmesh


# #############################################################################
# #############################################################################
#                           Mesh2DTri
# #############################################################################


def _mesh2DTri_conformity(knots=None, cents=None, key=None):

    # ---------------------------------
    # make sure np.ndarrays of dim = 2

    knots = np.atleast_2d(knots).astype(float)
    cents = np.atleast_2d(cents).astype(int)

    # --------------
    # check shapes

    c0 = (
        knots.shape[1] == 2
        and knots.shape[0] >= 3
        and cents.shape[1] in [3, 4]
        and cents.shape[0] >= 1
        and cents.dtype == np.int
    )
    if not c0:
        msg = (
            "Arg knots must be of shape (nknots>=3, 2) and "
            "arg cents must be of shape (ncents>=1, 3 or 4) and dtype = int\n"
            "Provided:\n"
            f"\t- knots.shape: {knots.shape}\n"
            f"\t- cents.shape: {cents.shape}\n"
            f"\t- cents.dtype: {cents.dtype}\n"
        )
        raise Exception(msg)

    nknots = knots.shape[0]
    ncents = cents.shape[0]

    # -------------------
    # Test for duplicates

    # knots (floats => distance)
    dist = np.full((nknots, nknots), np.nan)
    ind = np.zeros(dist.shape, dtype=bool)
    for ii in range(nknots):
        dist[ii, ii+1:] = np.sqrt(
            (knots[ii+1:, 0] - knots[ii, 0])**2
            + (knots[ii+1:, 1] - knots[ii, 1])**2
        )
        ind[ii, ii+1:] = True

    ind[ind] = dist[ind] < 1.e-6
    if np.any(ind):
        iind = np.any(ind, axis=1).nonzero()[0]
        lstr = [f'\t\t- {ii}: {ind[ii, :].nonzero()[0]}' for ii in iind]
        msg = (
            f"Non-valid mesh {key}: \n"
            f"  Duplicate knots: {ind.sum()}\n"
            f"\t- knots.shape: {cents.shape}\n"
            f"\t- duplicate indices:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # cents (indices)
    centsu = np.unique(cents, axis=0)
    if centsu.shape[0] != ncents:
        msg = (
            f"Non-valid mesh {key}: \n"
            f"  Duplicate cents: {ncents - centsu.shape[0]}\n"
            f"\t- cents.shape: {cents.shape}\n"
            f"\t- unique shape: {centsu.shape}"
        )
        raise Exception(msg)

    # ---------------------
    # Test for unused knots

    centsu = np.unique(centsu)
    c0 = np.all(centsu >= 0) and centsu.size == nknots
    if centsu.size < nknots:
        ino = (~np.in1d(
            range(0, nknots),
            centsu,
            assume_unique=False,
            invert=False,
        )).nonzero()[0]
        msg = (
            f"Unused knots in {key}:\n"
            f"\t- unused knots indices: {ino}"
        )
        warnings.warn(msg)
    elif centsu.size > nknots or centsu.max() != nknots - 1:
        unknown = np.setdiff1d(centsu, range(nknots), assume_sorted=True)
        msg = (
            "Unknown knots refered to in cents!\n"
            f"\t- unknown knots: {unknown}"
        )
        raise Exception(msg)

    return cents, knots


def _mesh2DTri_clockwise(knots=None, cents=None, key=None):

    x, y = knots[cents, 0], knots[cents, 1]
    orient = (
        (y[:, 1] - y[:, 0])*(x[:, 2] - x[:, 1])
        - (y[:, 2] - y[:, 1])*(x[:, 1] - x[:, 0])
    )

    clock = orient > 0.
    if np.any(clock):
        msg = (
            "Some triangles not counter-clockwise\n"
            "  (necessary for matplotlib.tri.Triangulation)\n"
            f"    => {clock.sum()}/{cents.shape[0]} triangles reshaped"
        )
        warnings.warn(msg)
        cents[clock, 1], cents[clock, 2] = cents[clock, 2], cents[clock, 1]
    return cents


def _mesh2DTri_to_dict(knots=None, cents=None, key=None, trifind=None):

    # ---------------------
    # check mesh conformity

    cents, knots = _mesh2DTri_conformity(knots=knots, cents=cents, key=key)

    # ---------------------------------------------
    # define triangular mesh and trifinder function

    # triangular mesh
    if cents.shape[1] == 3:

        # check clock-wise triangles
        cents = _mesh2DTri_clockwise(knots=knots, cents=cents, key=key)

        # check trifinder
        if trifind is None:
            mpltri = mplTri(knots[:, 0], knots[:, 1], cents)
            trifind = mpltri.get_trifinder()

        meshtype = 'tri'

    # Quadrangular mesh => convert to triangular
    elif cents.shape[1] == 4:

        cents2 = np.empty((cents.shape[0]*2, 3), dtype=int)
        cents2[::2, :] = cents[:, :3]
        cents2[1::2, :-1] = cents[:, 2:]
        cents2[1::2, -1] = cents[:, 0]
        cents = cents2

        # Re-check mesh conformity
        cents, knots = _mesh2DTri_conformity(knots=knots, cents=cents, key=key)
        cents = _mesh2DTri_clockwise(knots=knots, cents=cents, key=key)

        # check trifinder
        if trifind is None:
            mpltri = mplTri(knots[:, 0], knots[:, 1], cents)
            trifind = mpltri.get_trifinder()

        meshtype = 'quadtri'

    # ----------------------------
    # Check on trifinder function

    assert callable(trifind), "Arg trifind must be a callable!"

    try:
        out = trifind(np.r_[0.], np.r_[0])
        assert isinstance(out, np.ndarray)
    except Exception as err:
        msg = (
            "Arg trifind must return an array of indices when fed with arrays "
            "of (R, Z) coordinates!\n"
            f"\ttrifind(np.r_[0], np.r_[0.]) = {out}"
        )
        raise Exception(msg)

    # -----------------
    # Format ouput dict

    kcents = f"{key}-cents"
    kcentsR = f"{kcents}-R"
    kcentsZ = f"{kcents}-Z"
    kcents_pts = f"{kcents}-pts"
    kcents_ind = f"{kcents}-ind"
    kknots = f"{key}-knots"
    kknotsR = f"{kknots}-R"
    kknotsZ = f"{kknots}-Z"
    kknots_ind = f"{kknots}-ind"

    # dref
    dref = {
        kknots_ind: {
            'data': np.arange(0, knots.shape[0]),
            'units': '',
            # 'source': None,
            'dim': '',
            'quant': 'ind',
            'name': 'ind',
            'group': 'ind',
        },
        kcents_ind: {
            'data': np.arange(0, cents.shape[0]),
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'group': 'Z',
        },
        kcents_pts: {
            'data': np.arange(0, 3),
            'units': '',
            # 'source': None,
            'dim': '',
            'quant': 'ind',
            'name': 'ind',
            'group': 'ind',
        },
    }

    # ddata
    ddata = {
        kknotsR: {
            'data': knots[:, 0],
            'ref': (kknots_ind,),
            'units': 'm',
            'quant': 'R',
            'dim': 'distance',
            'group': 'R',
        },
        kknotsZ: {
            'data': knots[:, 1],
            'ref': (kknots_ind,),
            'units': 'm',
            'quant': 'Z',
            'dim': 'distance',
            'group': 'Z',
        },
        kcentsR: {
            'data': np.mean(knots[cents, 0], axis=1),
            'ref': (kcents_ind,),
            'units': 'm',
            'quant': 'R',
            'dim': 'distance',
            'group': 'R',
        },
        kcentsZ: {
            'data': np.mean(knots[cents, 1], axis=1),
            'ref': (kcents_ind,),
            'units': 'm',
            'quant': 'Z',
            'dim': 'distance',
            'group': 'Z',
        },
        kcents: {
            'data': cents,
            'ref': (kcents_ind, kcents_pts),
            'units': '',
            'quant': 'ind',
            'dim': 'ind',
            'group': 'ind',
        },
    }

    # dobj
    dmesh = {
        key: {
            'type': meshtype,
            'cents': kcents,
            'knots': kknots,
            'ref': (kcents,),
            'shape': (cents.shape[0],),
            'trifind': trifind,
            'crop': False,
        },
    }
    return dref, ddata, dmesh


# #############################################################################
# #############################################################################
#                           Mesh2DRect
# #############################################################################


def _mesh2DRect_X_check(
    x=None,
    res=None,
):
    """ Returns knots (x) and associated resolution

    res can be:
        - int: numbr of mesh elements desired between knots
        - float: desired average mesh element size
        - array of floats: (one for each x, desired approximate mesh size)

    """

    # ------------
    # Check inputs

    # x
    try:
        x = np.unique(np.ravel(x).astype(float))
    except Exception as err:
        msg = "x must be convertible to a sorted, flat array of floats!"
        raise Exception(msg)

    # res
    if res is None:
        res = 10

    lc = [
        isinstance(res, (int, np.int64, np.int32)) and len(x) == 2,
        isinstance(res, (float, np.floating)) and len(x) == 2,
        isinstance(res, (list, tuple, np.ndarray)) and len(x) == len(res),
    ]
    if not any(lc):
        msg = (
            "Arg res must be:\n"
            "\t- int: nb of mesh elements along x\n"
            "\t       requires len(x) = 2\n"
            "\t- float: approximate desired mesh element size along x\n"
            "\t       requires len(x) = 2\n"
            "\t- iterable: approximate desired mesh element sizes along x\n"
            "\t       requires len(x) = len(res)\n"
        )
        raise Exception(msg)

    if lc[0]:
        x_new = np.linspace(x[0], x[1], int(res)+1)
        res_new = res
        indsep = None

    elif lc[1]:
        nb = int(np.ceil((x[1]-x[0]) / res))
        x_new = np.linspace(x[0], x[1], nb+1)
        res_new = np.mean(np.diff(x))
        indsep = None

    else:

        # check conformity
        res = np.ravel(res).astype(float)
        delta = np.diff(x)
        res_sum = res[:-1] + res[1:]
        ind = res_sum > delta + 1.e-14
        if np.any(ind):
            msg = (
                "Desired resolution is not achievable for the following:\n"
                f"res_sum: {res_sum[ind]}\n"
                f"delta  : {delta[ind]}"
            )
            raise Exception(msg)

        # compute nn
        # nn = how many pairs can fit in the interval
        npairs = np.round(delta/res_sum).astype(int)
        res_sum_new = delta / npairs

        fract = res[:-1] / res_sum

        res_new = [None for ii in range(len(x)-1)]
        x_new = [None for ii in range(len(x)-1)]
        for ii in range(len(x)-1):
            res_new[ii] = (
                res_sum_new[ii]
                * np.linspace(fract[ii], 1.-fract[ii], 2*npairs[ii])
            )
            if ii == 0:
                res_add = np.concatenate(([0], np.cumsum(res_new[ii])))
            else:
                res_add = np.cumsum(res_new[ii])
            x_new[ii] = x[ii] + res_add

        indsep = np.cumsum(npairs[:-1]*2)
        res_new = np.concatenate(res_new)
        x_new = np.concatenate(x_new)

    return x_new, res_new, indsep


def _mesh2DRect_check(
    domain=None,
    res=None,
    R=None,
    Z=None,
):

    # --------------
    # check inputs

    # (domain, res) vs (R, Z)
    lc = [
        domain is not None,
        R is not None and Z is not None,
    ]
    if all(lc) or not any(lc):
        msg = (
            "Please provide (domain, res) xor (R, Z), not both:\n"
            "Provided:\n"
            f"\t- domain, res: {domain}, {res}\n"
            f"\t- R, Z: {R}, {Z}\n"
        )
        raise Exception(msg)

    if lc[0]:
        # domain
        c0 = (
            isinstance(domain, list)
            and len(domain) == 2
            and all([
                hasattr(dd, '__iter__') and len(dd) >= 2 for dd in domain
            ])
        )
        if not c0:
            msg = (
                "Arg domain must be a list of 2 iterables of len() >= 2\n"
                f"Provided: {domain}"
            )
            raise Exception(msg)

        # res
        c0 = (
            res is None
            or np.isscalar(res)
            or isinstance(res, list) and len(res) == 2
        )
        if not c0:
            msg = (
                "Arg res must be a int, float or array or a list of 2 such\n"
                f"Provided: {res}"
            )
            raise Exception(msg)

        if np.isscalar(res) or res is None:
            res = [res, res]

        # -------------
        # check R and Z

        R, resR, indR = _mesh2DRect_X_check(domain[0], res=res[0])
        Z, resZ, indZ = _mesh2DRect_X_check(domain[1], res=res[1])

    elif lc[1]:

        # R, Z check
        c0 = (
            hasattr(R, '__iter__')
            and np.asarray(R).ndim == 1
            and np.unique(R).size == np.array(R).size
            and np.allclose(np.unique(R), R)
        )
        if not c0:
            msg = "Arg R must be convertible to a 1d increasing array"
            raise Exception(msg)

        c0 = (
            hasattr(Z, '__iter__')
            and np.asarray(Z).ndim == 1
            and np.unique(Z).size == np.array(Z).size
            and np.allclose(np.unique(Z), Z)
        )
        if not c0:
            msg = "Arg Z must be convertible to a 1d increasing array"
            raise Exception(msg)

        R = np.unique(R)
        Z = np.unique(Z)

        resR = np.diff(R)
        resZ = np.diff(Z)

        if np.unique(resR).size == 1:
            resR = resR[0]
            indR = None
        else:
            raise NotImplementedError()
        if np.unique(resZ).size == 1:
            resZ = resZ[0]
            indZ = None
        else:
            raise NotImplementedError()

    return R, Z, resR, resZ, indR, indZ


def _mesh2DRect_to_dict(
    domain=None,
    res=None,
    R=None,
    Z=None,
    key=None,
):

    # --------------------
    # check / format input

    kRknots, kZknots = f"{key}-R-knots", f"{key}-Z-knots"
    kRcent, kZcent = f"{key}-R-cents", f"{key}-Z-cents"

    R, Z, resR, resZ, indR, indZ = _mesh2DRect_check(
        domain=domain,
        res=res,
        R=R,
        Z=Z,
    )
    Rcent = 0.5*(R[1:] + R[:-1])
    Zcent = 0.5*(Z[1:] + Z[:-1])

    variable = not (np.isscalar(resR) and np.isscalar(resZ))

    # --------------------
    # prepare dict

    dref = {
        kRknots: {
            'data': R,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'group': 'R',
        },
        kZknots: {
            'data': Z,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'group': 'Z',
        },
        kRcent: {
            'data': Rcent,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'group': 'R',
        },
        kZcent: {
            'data': Zcent,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'group': 'Z',
        },
    }

    # dobj
    dmesh = {
        key: {
            'type': 'rect',
            'knots': (kRknots, kZknots),
            'cents': (kRcent, kZcent),
            'ref': (kRcent, kZcent),
            'shape': (Rcent.size, Zcent.size),
            'variable': variable,
            'crop': False,
        },
    }
    return dref, dmesh


def _mesh2DRect_from_croppoly(crop_poly=None, domain=None):

    # ------------
    # check inputs

    c0 = hasattr(crop_poly, '__iter__') and len(crop_poly) == 2
    lc = [
        crop_poly is None,
        (
            c0
            and isinstance(crop_poly, tuple)
            and crop_poly[0].__class__.__name__ == 'Config'
            and (isinstance(crop_poly[1], str) or crop_poly[1] is None)
        )
        or crop_poly.__class__.__name__ == 'Config',
        c0
        and all([
            hasattr(cc, '__iter__') and len(cc) == len(crop_poly[0])
            for cc in crop_poly[1:]
        ])
        and np.asarray(crop_poly).ndim == 2
    ]

    if not any(lc):
        msg = (
            "Arg config must be a Config instance!"
        )
        raise Exception(msg)

    # -------------
    # Get polyand domain

    if lc[0]:
        # trivial case
        poly = None

    else:

        # -------------
        # Get poly from input

        if lc[1]:
            # (config, structure name)

            if crop_poly.__class__.__name__ == 'Config':
                config = crop_poly
                key_struct = None
            else:
                config, key_struct = crop_poly

            # key_struct if None
            if key_struct is None:
                lk, ls = zip(*[
                    (ss.Id.Name, ss.dgeom['Surf']) for ss in config.lStructIn
                ])
                key_struct = lk[np.argmin(ls)]

            # poly
            poly = config.dStruct['dObj']['Ves'][key_struct].Poly_closed

        else:

            # make sure poly is np.ndarraya and closed
            poly = np.asarray(crop_poly).astype(float)
            if not np.allclose(poly[:, 0], poly[:, -1]):
                poly = np.concatenate((poly, poly[:, 0:1]))

        # -------------
        # Get domain from poly

        if domain is None:
            domain = [
                [poly[0, :].min(), poly[0, :].max()],
                [poly[1, :].min(), poly[1, :].max()],
            ]

    return domain, poly


# #############################################################################
# #############################################################################
#                           Mesh2DRect - select
# #############################################################################


def _select_ind_check(
    ind=None,
    elements=None,
    returnas=None,
    crop=None,
    meshtype=None,
):

    # ----------------------
    # check basic conditions

    if meshtype == 'rect':
        lc = [
            ind is None,
            isinstance(ind, tuple)
            and len(ind) == 2
            and (
                all([np.isscalar(ss) for ss in ind])
                or all([
                    hasattr(ss, '__iter__')
                    and len(ss) == len(ind[0])
                    for ss in ind
                ])
                or all([isinstance(ss, np.ndarray) for ss in ind])
            ),
            (
                np.isscalar(ind)
                or (
                    hasattr(ind, '__iter__')
                    and all([np.isscalar(ss) for ss in ind])
                )
                or isinstance(ind, np.ndarray)
            )
        ]
    else:
        lc = [
            ind is None,
            np.isscalar(ind)
            or (
                hasattr(ind, '__iter__')
                and all([np.isscalar(ss) for ss in ind])
            )
            or isinstance(ind, np.ndarray)
        ]

    if not any(lc):
        if meshtype == 'rect':
            msg = (
                "Arg ind must be either:\n"
                "\t- None\n"
                "\t- int or array of int: int indices in mixed (R, Z) index\n"
                "\t- tuple of such: int indices in (R, Z) index respectively\n"
                f"Provided: {ind}"
            )
        else:
            msg = (
                "Arg ind must be either:\n"
                "\t- None\n"
                "\t- int or array of int: int indices\n"
                "\t- array of bool: bool indices\n"
                f"Provided: {ind}"
            )
        raise Exception(msg)

    # ----------------------
    # adapt to each case

    if lc[0]:
        pass
    elif lc[1] and meshtype == 'rect':
        if any([not isinstance(ss, np.ndarray) for ss in ind]):
            ind = (
                np.atleast_1d(ind[0]).astype(int),
                np.atleast_1d(ind[1]).astype(int),
            )
        lc0 = [
            [
                isinstance(ss, np.ndarray),
                np.issubdtype(ss.dtype, np.integer),
                ss.shape == ind[0].shape,
            ]
                for ss in ind
        ]
        if not all([all(cc) for cc in lc0]):
            ltype = [type(ss) for ss in ind]
            ltypes = [
                ss.dtype if isinstance(ss, np.ndarray) else False
                for ss in ind
            ]
            lshapes = [
                ss.shape if isinstance(ss, np.ndarray) else len(ss)
                for ss in ind
            ]
            msg = (
                "Arg ind must be a tuple of 2 arrays of int of same shape\n"
                f"\t- lc0: {lc0}\n"
                f"\t- types: {ltype}\n"
                f"\t- type each: {ltypes}\n"
                f"\t- shape: {lshapes}\n"
                f"\t- ind: {ind}"
            )
            raise Exception(msg)
    elif lc[1] and meshtype in ['tri', 'quadtri']:
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                "Arg ind must be an array of bool or int\n"
                f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    else:
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                "Arg ind must be an array of bool or int\n"
                f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    # elements
    elements = _generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    if meshtype == 'rect':
        retdef = tuple
        retok = [tuple, np.ndarray, 'tuple-flat', 'array-flat', bool]
    else:
        retdef = bool
        retok = [int, bool]
    returnas = _generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default=retdef,
        allowed=retok,
    )

    # crop
    crop = _generic_check._check_var(
        crop, 'crop',
        types=bool,
        default=True,
    )

    return ind, elements, returnas, crop


def _select_check(
    elements=None,
    returnas=None,
    return_ind_as=None,
    return_neighbours=None,
):

    # elements
    elements = _generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    returnas = _generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default='ind',
        allowed=['ind', 'data'],
    )

    # return_ind_as
    return_ind_as = _generic_check._check_var(
        return_ind_as, 'return_ind_as',
        types=None,
        default=int,
        allowed=[int, bool],
    )

    # return_neighbours
    return_neighbours = _generic_check._check_var(
        return_neighbours, 'return_neighbours',
        types=bool,
        default=True,
    )

    return elements, returnas, return_ind_as, return_neighbours,


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _mesh2D_bsplines(key=None, lkeys=None, deg=None):

    # key
    key = _generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lkeys,
    )

    # deg
    deg = _generic_check._check_var(
        deg, 'deg',
        types=int,
        default=2,
        allowed=[0, 1, 2, 3],
    )

    # keybs
    keybs = f'{key}-bs{deg}'

    return key, keybs, deg



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

# tofu
# from tofu import __version__ as __version__
from . import _generic_check


# #############################################################################
# #############################################################################
#                           inversions
# #############################################################################


def _plot_geometry_matrix_check(
    coll=None,
    key=None,
    indbf=None,
    indchan=None,
    cmap=None,
    dcolorbar=None,
    dleg=None,
    dax=None,
):

    # key
    if 'inversions' not in coll.dobj.keys():
        msg = 'No inversions available!'
        raise Exception(msg)

    lk = list(coll.dobj['inversions'].keys())
    keyinv = _generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lk,
    )
    keymat = coll.dobj['inversions'][keyinv]['matrix']
    keydata = coll.dobj['inversions'][keyinv]['data_in']
    keybs = coll.dobj['matrix'][keymat]['bsplines']
    # refbs = coll.dobj['bsplines'][keybs]['ref']

    crop = coll.dobj['matrix'][keymat]['crop']
    if crop is True:
        cropbs = coll.dobj['bsplines'][keybs]['crop']
        cropbs = coll.ddata[cropbs]['data']
    else:
        cropbs = None

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = _generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = _generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return keyinv, keymat, keybs, keydata, cropbs, cmap, dcolorbar, dleg


def plot_inversion(
    coll=None,
    key=None,
    indt=None,
    res=None,
    vmin=None,
    vmax=None,
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
):

    # ------------
    # check inputs

    (
        keyinv, keymat, keybs, keydata, cropbs, cmap, dcolorbar, dleg,
    ) = _plot_geometry_matrix_check(
        coll=coll,
        key=key,
        cmap=cmap,
        dcolorbar=dcolorbar,
        dleg=dleg,
    )

    # ------------
    # prepare data

    # synthetic ?
    synthetic = keydata in coll.dobj.get('synthetic', {}).keys()

    # data
    keychan = coll.ddata[keydata]['ref'][-1]
    chan = coll.ddata[keychan]['data']
    if 'time' in coll.ddata[keyinv]['group']:
        data = coll.ddata[keydata]['data']
        keyt = coll.ddata[keydata]['ref'][0]
        time = coll.ddata[keyt]['data']
    else:
        data = coll.ddata[keydata]['data'][None, :]
        time = [0]
        indt = 0

    # reconstructed data
    matrix = coll.ddata[keymat]['data']
    sol = coll.ddata[keyinv]['data']
    shapebs = coll.dobj['bsplines'][keybs]['shape']
    nbs = int(np.prod(shapebs))

    if cropbs is not None:
        cropbs_flat = cropbs.ravel(order='F')
    if 'time' in coll.ddata[keyinv]['group']:
        nt = sol.shape[0]
        if sol.ndim == 3:
            sol_flat = sol.reshape((nt, nbs), order='F')
        else:
            sol_flat = sol
        if cropbs is not None:
            data_re = matrix.dot(sol_flat[:, cropbs_flat].T)
        else:
            data_re = matrix.dot(sol_flat[:, ...].T)
    else:
        if sol.ndim == 2:
            sol_flat = sol.ravel(order='F')
        else:
            sol_flat = sol
        if cropbs is not None:
            data_re = matrix.dot(sol_flat[cropbs_flat].T)[:, None]
        else:
            data_re = matrix.dot(sol_flat.T)[:, None]

    # inversion parameters
    if 'time' in coll.ddata[keyinv]['group']:
        nchi2n = chan.size * coll.ddata[f'{keyinv}-niter']['data']
        mu = coll.ddata[f'{keyinv}-mu']['data']
        reg = coll.ddata[f'{keyinv}-reg']['data']
        niter = coll.ddata[f'{keyinv}-niter']['data']
    else:
        nchi2n = chan.size * coll.dobj['inversions'][keyinv]['chi2n']
        mu = coll.dobj['inversions'][keyinv]['mu']
        reg = coll.dobj['inversions'][keyinv]['reg']
        niter = coll.dobj['inversions'][keyinv]['niter']

    # indt
    if indt is None:
        indt = 0

    # --------------
    # plot - prepare

    if dax is None:

        if fs is None:
            fs = (16, 9)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.98,
                'bottom': 0.05, 'top': 0.95,
                'hspace': 0.15, 'wspace': 0.15,
            }

        fig = plt.figure(figsize=fs)
        if synthetic:
            gs = gridspec.GridSpec(ncols=4, nrows=2, **dmargin)
            ax0 = fig.add_subplot(gs[1, 1])
            ax1 = fig.add_subplot(gs[:2, 2:])
            ax2 = fig.add_subplot(gs[2, 2:], sharex=ax1)
            ax3 = fig.add_subplot(gs[3, 2:])
            ax4 = fig.add_subplot(gs[4, 2:], sharex=ax3)

            ax5 = fig.add_subplot(gs[0, 1], sharex=ax0, sharey=ax0)
            ax6 = fig.add_subplot(gs[1, 0], sharex=ax0, sharey=ax0)

        else:
            gs = gridspec.GridSpec(ncols=2, nrows=5, **dmargin)
            ax0 = fig.add_subplot(gs[:, 0])
            ax1 = fig.add_subplot(gs[:2, 1])
            ax2 = fig.add_subplot(gs[2, 1], sharex=ax1)
            ax3 = fig.add_subplot(gs[3, 1])
            ax4 = fig.add_subplot(gs[4, 1], sharex=ax3)

        ax0.set_xlabel(f'R (m)')
        ax0.set_ylabel(f'Z (m)')
        ax0.set_title('reconstruction', size=14)

        ax1.set_xlabel(f'chan')
        ax1.set_ylabel(f'data')
        ax1.set_title('data', size=14)

        ax2.set_xlabel(f'chan')
        ax2.set_ylabel(f'error')

        ax3.set_xlabel(f'time')
        ax3.set_ylabel(f'potential')

        ax4.set_xlabel(f'time')
        ax4.set_ylabel(f'niter')

        dax = {
            'reconstruction': {'ax': ax0, 'type': 'cross'},
            'data': {'ax': ax1, 'type': 'misc'},
            'data-err': {'ax': ax2, 'type': 'misc'},
            'inv-param': {'ax': ax3, 'type': 'misc'},
            'niter': {'ax': ax4, 'type': 'misc'},
        }

        if synthetic:
            ax5.set_xlabel(f'R (m)')
            ax5.set_ylabel(f'Z (m)')
            ax5.set_title('original', size=14)

            ax6.set_xlabel(f'R (m)')
            ax6.set_ylabel(f'Z (m)')
            ax6.set_title('error', size=14)

            dax.update({
                'original': {'ax': ax5, 'type': 'cross'},
                'error': {'ax': ax6, 'type': 'cross'},
            })

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # ---------
    # plot inv

    kax = 'reconstruction'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']
        coll.plot_profile2d(
            key=keyinv,
            indt=indt,
            res=res,
            dax={'cross': ax},
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            dleg=False,
        )

    # ---------
    # plot data

    kax = 'data'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']
        ax.plot(
            chan,
            data[indt, :],
            c='k',
            ls='-',
            lw=1.,
            marker='.',
        )
        ax.plot(
            chan,
            data_re[:, indt],
            c='k',
            ls='--',
            lw=1.,
            marker='.',
        )

    kax = 'data-err'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']
        ax.plot(
            chan,
            data_re[:, indt] - data[indt, :],
            c='k',
            ls='-',
            lw=1.,
            marker='.',
        )
        ax.axhline(
            0.,
            c='k',
            ls='-',
            lw=1.,
        )

    # -------------------------
    # plot inversion parameters

    kax = 'inv-param'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']
        ax.plot(
            time,
            nchi2n + mu*reg,
            c='k',
            ls='-',
            lw=1.,
            marker='.',
            label='n*chi2n + mu*reg',
        )
        ax.plot(
            time,
            nchi2n,
            c='r',
            ls='-',
            lw=1.,
            marker='.',
            label='nchi2n',
        )
        ax.plot(
            time,
            mu*reg,
            c='b',
            ls='-',
            lw=1.,
            marker='.',
            label='mu*reg',
        )
        ax.axvline(time[indt], c='k', ls='-', lw=1.)

    kax = 'niter'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']
        ax.plot(
            time,
            niter,
            c='k',
            ls='-',
            lw=1.,
            marker='.',
        )
        ax.axvline(time[indt], c='k', ls='-', lw=1.)

    return dax



# -*- coding: utf-8 -*-


# Built-in
import datetime as dtm


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.colors as mcolors


# specific
from . import _generic_check


# #############################################################################
# #############################################################################
#                           plot matrix
# #############################################################################


def _plot_geometry_matrix_check(
    coll=None,
    key=None,
    indbf=None,
    indchan=None,
    cmap=None,
    aspect=None,
    dcolorbar=None,
    dleg=None,
    dax=None,
):

    # key
    lk = list(coll.dobj['matrix'].keys())
    key = _generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lk,
    )
    keybs = coll.dobj['matrix'][key]['bsplines']
    refbs = coll.dobj['bsplines'][keybs]['ref']
    keym = coll.dobj['bsplines'][keybs]['mesh']

    # indbf
    if indbf is None:
        indbf = 0
    try:
        assert np.isscalar(indbf)
        indbf = int(indbf)
    except Exception as err:
        msg = (
            f"Arg indbf should be a int!\nProvided: {indt}"
        )
        raise Exception(msg)

    # indchan
    if indchan is None:
        indchan = 0
    try:
        assert np.isscalar(indchan)
        indchan = int(indchan)
    except Exception as err:
        msg = (
            f"Arg indchan should be a int!\nProvided: {indt}"
        )
        raise Exception(msg)

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # aspect
    aspect = _generic_check._check_var(
        aspect, 'aspect',
        default='auto',
        types=str,
        allowed=['auto', 'equal'],
    )

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = _generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = _generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return key, keybs, keym, indbf, indchan, cmap, aspect, dcolorbar, dleg


def _plot_geometry_matrix_prepare(
    cam=None,
    coll=None,
    key=None,
    keybs=None,
    keym=None,
    indbf=None,
    indchan=None,
    res=None,
):

    # res
    deg = coll.dobj['bsplines'][keybs]['deg']
    km = coll.dobj['bsplines'][keybs]['mesh']
    kR, kZ = coll.dobj['mesh'][km]['knots']
    Rk = coll.ddata[kR]['data']
    Zk = coll.ddata[kZ]['data']
    dR = np.min(np.diff(Rk))
    dZ = np.min(np.diff(Zk))
    if res is None:
        res_coef = 0.025
        res = [res_coef*dR, res_coef*dZ]

    # crop
    crop = coll.dobj['matrix'][key]['crop']

    # indchan => indchan_bf
    ich_bf_tup = coll.select_ind(key=keybs, returnas='tuple-flat', crop=crop)
    ich_bf = coll.select_ind(key=keybs, returnas=np.ndarray, crop=crop)
    indbf_full = coll.select_ind(
        key=keybs, returnas='array-flat', crop=crop,
    )[indbf]
    indbf_tup = coll.select_ind(
        key=keybs, ind=indbf_full, returnas=tuple, crop=crop,
    )

    # mesh sampling
    km = coll.dobj['bsplines'][keybs]['mesh']
    R, Z = coll.get_sample_mesh(
        key=km, res=res, mode='abs', grid=True, imshow=True,
    )

    # bsplinetot
    shapebs = coll.dobj['bsplines'][keybs]['shape']
    coefs = np.zeros(tuple(np.r_[1, shapebs]), dtype=float)
    coefs[0, ich_bf_tup[0], ich_bf_tup[1]] = np.nansum(
        coll.ddata[key]['data'],
        axis=0,
    )

    bsplinetot = coll.interp2d(
        key=keybs,
        R=R,
        Z=Z,
        coefs=coefs,
        crop=crop,
        nan0=True,
        details=False,
    )[0, ...]

    # bspline1
    coefs[0, ich_bf_tup[0], ich_bf_tup[1]] = (
        coll.ddata[key]['data'][indchan, :]
    )
    bspline1 = coll.interp2d(
        key=keybs,
        R=R,
        Z=Z,
        coefs=coefs,
        crop=crop,
        nan0=True,
        details=False,
    )[0, ...]

    # los
    ptslos, coefslines, indlosok = None, None, None
    if cam is not None:
        ptslos = cam._get_plotL(return_pts=True, proj='cross', Lplot='tot')
        indsep = np.nonzero(np.isnan(ptslos[0, :]))[0]
        ptslos = np.split(ptslos, indsep, axis=1)
        coefslines = coll.ddata[key]['data'][:, indbf]
        indlosok = np.nonzero(coefslines > 0)[0]
        # normalize for line width
        coefslines = (
            (3. - 0.5) * (coefslines - coefslines.min())
            / (coefslines.max() - coefslines.min()) + 0.5
        )

    # extent and interp
    extent = (
        Rk[0] - 0.*dR, Rk[-1] + 0.*dR,
        Zk[0] - 0.*dZ, Zk[-1] + 0.*dZ,
    )

    if deg == 0:
        interp = 'nearest'
    elif deg == 1:
        interp = 'bilinear'
    elif deg >= 2:
        interp = 'bicubic'

    return (
        bsplinetot, bspline1, extent, interp,
        ptslos, coefslines, indlosok, indbf_tup,
    )


def plot_geometry_matrix(
    cam=None,
    coll=None,
    key=None,
    indbf=None,
    indchan=None,
    vmin=None,
    vmax=None,
    res=None,
    cmap=None,
    aspect=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
):

    # --------------
    # check input

    (
        key, keybs, keym,
        indbf, indchan,
        cmap, aspect, dcolorbar, dleg,
    ) = _plot_geometry_matrix_check(
        coll=coll,
        key=key,
        indbf=indbf,
        indchan=indchan,
        cmap=cmap,
        aspect=aspect,
        dcolorbar=dcolorbar,
        dleg=dleg,
        dax=dax,
    )

    # --------------
    #  Prepare data

    (
        bsplinetot, bspline1,
        extent, interp,
        ptslos, coefslines, indlosok,
        ich_bf_tup,
    ) = _plot_geometry_matrix_prepare(
        cam=cam,
        coll=coll,
        key=key,
        keybs=keybs,
        keym=keym,
        indbf=indbf,
        indchan=indchan,
        res=res,
    )
    nchan, nbs = coll.ddata[key]['data'].shape

    # --------------
    # plot - prepare

    if dax is None:

        if fs is None:
            fs = (16, 9)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.98,
                'bottom': 0.05, 'top': 0.95,
                'hspace': 0.15, 'wspace': 0.15,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=3, nrows=2, **dmargin)
        ax01 = fig.add_subplot(gs[0, 1])
        ax01.set_ylabel(f'channels')
        ax01.set_xlabel(f'basis functions')
        ax01.set_title(key, size=14)
        ax01.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax01.xaxis.set_label_position('top')
        ax00 = fig.add_subplot(gs[0, 0], sharex=ax01)
        ax00.set_xlabel(f'basis functions (m)')
        ax00.set_ylabel(f'data')
        ax10 = fig.add_subplot(gs[1, 0], aspect='equal')
        ax10.set_xlabel(f'R (m)')
        ax10.set_ylabel(f'Z (m)')
        ax11 = fig.add_subplot(gs[1, 1], aspect='equal')
        ax11.set_ylabel(f'R (m)')
        ax11.set_xlabel(f'Z (m)')
        ax02 = fig.add_subplot(gs[0, 2], sharey=ax01)
        ax02.set_xlabel(f'channels')
        ax02.set_ylabel(f'data')
        ax02.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax02.xaxis.set_label_position('top')
        ax02.tick_params(
            axis="y",
            left=False, right=True,
            labelleft=False, labelright=True,
        )
        ax02.yaxis.set_label_position('right')
        ax12 = fig.add_subplot(gs[1, 2], aspect='equal')
        ax12.set_xlabel(f'R (m)')
        ax12.set_ylabel(f'Z (m)')

        dax = {
            'matrix': ax01,
            'cross1': {'ax': ax10, 'type': 'cross'},
            'cross2': {'ax': ax12, 'type': 'cross'},
            'crosstot': {'ax': ax11, 'type': 'cross'},
            'misc1': {'ax': ax02, 'type': 'misc'},
            'misc2': {'ax': ax00, 'type': 'misc'},
        }

    dax = _generic_check._check_dax(dax=dax, main='matrix')

    # --------------
    # plot mesh

    dax = coll.plot_mesh(
        key=keym, dax=dax, crop=True, dleg=False,
    )

    # --------------
    # plot matrix

    dax = coll.plot_as_matrix(
        key=key, dax=dax, ind=[indchan, indbf],
        cmap=cmap, vmin=vmin, vmax=vmax, aspect=aspect,
    )

    kax = 'cross1'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']

        im = ax.imshow(
            bspline1,
            extent=extent,
            interpolation=interp,
            origin='lower',
            aspect='equal',
            cmap=cmap,
            vmin=0,
            vmax=None,
        )

        if ptslos is not None:
            ax.plot(
                ptslos[indchan][0, :],
                ptslos[indchan][1, :],
                ls='-',
                lw=1.,
                color='k',
            )

    kax = 'cross2'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']

        coll.plot_bsplines(
            key=keybs,
            ind=ich_bf_tup,
            knots=False,
            cents=False,
            plot_mesh=False,
            dax={'cross': dax[kax]},
            dleg=False,
        )

        if ptslos is not None:
            for ii in indlosok:
                ax.plot(
                    ptslos[ii][0, :],
                    ptslos[ii][1, :],
                    ls='-',
                    lw=coefslines[ii],
                    color='k',
                )

    kax = 'crosstot'
    if dax.get(kax) is not None:
        ax = dax[kax]['ax']

        im = ax.imshow(
            bsplinetot,
            extent=extent,
            interpolation=interp,
            origin='lower',
            aspect='equal',
            cmap=cmap,
            vmin=0,
            vmax=None,
        )

    # --------------
    # dleg

    # if dleg is not False:
        # dax['cross'].legend(**dleg)

    return dax



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import shutil
import itertools as itt
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu as tf
import tofu.data as tfd


_HERE = os.path.abspath(os.path.dirname(__file__))
_PATH_DATA = os.path.join(_HERE, 'test_data')
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_HERE)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')
VerbHead = 'tofu.mesh.test_01_checks'


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     checking routines
#
#######################################################


class Test01_checks():

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def setup(self):
        pass

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_mesh2DRect_X_check(self):

        lx = [[1, 2], [1, 2, 3, 4]]
        lres = [None, 10, 0.1, [0.1, 0.2], [0.1, 0.2, 0.3, 0.1]]

        for comb in itt.product(lx, lres):
            if hasattr(lres, '__iter__') and len(lres) != len(lx):
                continue
            x, res, ind = tfd._mesh_checks._mesh2DRect_X_check(
                x=[1, 2, 3, 4],
                res=10,
            )
            if hasattr(lres, '__iter__'):
                assert x_new.size == np.unique(x_new).size == res.size + 1


#######################################################
#
#     object mesh2D
#
#######################################################


class Test02_Mesh2D():

    @classmethod
    def setup_class(cls):
        pass

    def setup(self):
        self.dobj = {
            'm0': tfd.Mesh2D(),
            'm1': tfd.Mesh2D(),
            'm2': tfd.Mesh2D(),
            'm3': tfd.Mesh2D(),
        }

        # add mesh
        ldomain = [
            [[2, 3], [-1, 1]],
            [[2, 2.3, 2.6, 3], [-1, 0., 1]],
            [[2, 3], [-1, 0, 1]],
        ]
        lres = [
            0.1,
            [[0.2, 0.1, 0.1, 0.2], [0.2, 0.1, 0.2]],
            [0.1, [0.2, 0.1, 0.2]],
        ]

        i0 = 0
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            if k0 != 'm2':
                self.dobj[k0].add_mesh(
                    domain=ldomain[i0],
                    res=lres[i0],
                    key=k0,
                )
                i0 += 1
            else:
                self.dobj[k0].add_mesh(
                    crop_poly=tf.load_config('WEST'),
                    res=0.1,
                    key=k0,
                )

        # add splines
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            self.dobj[k0].add_bsplines(deg=ii)

        # Add triangular mesh
        knots = np.array([
            [2, 0], [2, 1], [3, 0], [3, 1],
        ])
        faces = np.array([[0, 1, 2], [1, 2, 3]])
        self.dobjtri = {
            'tri0': tf.data.Mesh2D(),
            'tri1': tf.data.Mesh2D(),
        }
        self.dobjtri['tri0'].add_mesh(cents=faces, knots=knots, key='tri0')

        # Add realistic NICE mesh for WEST
        pfe = os.path.join(_PATH_DATA, 'mesh_triangular_WEST_eq.txt')
        out = np.loadtxt(pfe)
        nknots, ncents = int(out[0, 0]), int(out[0, 1])
        assert out.shape == (nknots + ncents + 1, 3)
        knots = out[1:nknots + 1, :][:, :2]
        cents = out[nknots + 1:, :]
        self.dobjtri['tri1'].add_mesh(cents=cents, knots=knots, key='tri1')

        # add splines
        for ii, (k0, v0) in enumerate(self.dobjtri.items()):
            self.dobjtri[k0].add_bsplines(deg=ii)


    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_get_summary(self):
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            self.dobj[k0].get_summary()
        for ii, (k0, v0) in enumerate(self.dobjtri.items()):
            self.dobjtri[k0].get_summary()

    def test02_select_ind(self):

        # Rect mesh
        lkey = ['m0', 'm1-bs1', 'm2', 'm3-bs3']
        lelements = ['knots', None, 'cents', None]
        lind = [None, ([0, 5], [0, 6]), [0, 10, 100], ([0, 5, 6], [0, 2, 3])]
        lcrop = [True, False, True, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            indt = self.dobj[k0].select_ind(
                key=lkey[ii],
                ind=lind[ii],
                elements=lelements[ii],
                returnas=tuple,
                crop=lcrop[ii],
            )
            indf = self.dobj[k0].select_ind(
                key=lkey[ii],
                ind=indt,
                elements=lelements[ii],
                returnas=np.ndarray,
                crop=lcrop[ii],
            )
            indt2 = self.dobj[k0].select_ind(
                key=lkey[ii],
                ind=indf,
                elements=lelements[ii],
                returnas=tuple,
                crop=lcrop[ii],
            )
            assert all([np.allclose(indt[ii], indt2[ii]) for ii in [0, 1]])

        # triangular meshes
        lkeys = ['tri0', 'tri0', 'tri1']
        lind = [None, [1], 1]
        lelements = ['knots', None, 'cents']
        for ii, k0 in enumerate(lkeys):
            out = self.dobjtri[k0].select_ind(
                key=k0,
                ind=lind[ii],
                elements=lelements[ii],
                returnas=int,
                crop=lcrop[ii],
            )
            if ii == 0:
                assert np.allclose(out, np.r_[0, 1, 2, 3])
            elif ii >= 1:
                assert np.allclose(out, np.r_[1])

    def test03_select_mesh(self):

        # rectangular meshes
        lkey = ['m0', 'm1', 'm2', 'm3']
        lind = [None, ([0, 5], [0, 6]), [0, 10, 100], ([0, 5, 6], [0, 2, 3])]
        lelements = ['cents', 'knots', 'cents', None]
        lreturnas = ['ind', 'data', 'data', 'ind']
        lreturn_neig = [None, True, False, True]
        lcrop = [False, True, True, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            indf = self.dobj[k0].select_mesh_elements(
                key=lkey[ii],
                ind=lind[ii],
                elements=lelements[ii],
                returnas=lreturnas[ii],
                return_neighbours=lreturn_neig[ii],
                crop=lcrop[ii],
            )

        # triangular meshes
        lkeys = ['tri0', 'tri0', 'tri0', 'tri1']
        lind = [None, [1], 1, [0, 1]]
        lelements = ['knots', None, 'cents', 'cents']
        lreturnas = ['ind', 'data', 'ind', 'data']
        for ii, k0 in enumerate(lkeys):
            out = self.dobjtri[k0].select_mesh_elements(
                key=k0,
                ind=lind[ii],
                elements=lelements[ii],
                returnas=lreturnas[ii],
                return_neighbours=True,
                crop=lcrop[ii],
            )

    def test04_select_bsplines(self):

        # rectangular meshes
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        lind = [None, ([0, 5], [0, 6]), [0, 10, 100], ([0, 5, 6], [0, 2, 3])]
        lreturnas = [None, 'data', 'data', 'ind']
        lreturn_cents = [None, True, False, True]
        lreturn_knots = [None, False, True, True]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            indf = self.dobj[k0].select_bsplines(
                key=lkey[ii],
                ind=lind[ii],
                returnas=lreturnas[ii],
                return_cents=lreturn_cents[ii],
                return_knots=lreturn_knots[ii],
            )

        # triangular meshes
        lkeys = ['tri0', 'tri0', 'tri0', 'tri1']
        lkeysbs = ['tri0-bs0', None, 'tri0-bs0', 'tri1-bs1']
        lind = [None, [1], 1, [0, 1]]
        lelements = ['knots', None, 'cents', 'cents']
        lreturnas = ['ind', 'data', 'ind', 'data']
        for ii, k0 in enumerate(lkeys):
            indf = self.dobjtri[k0].select_bsplines(
                key=lkeysbs[ii],
                ind=lind[ii],
                returnas=lreturnas[ii],
                return_cents=lreturn_cents[ii],
                return_knots=lreturn_knots[ii],
            )

    def test05_sample_mesh(self):

        # rectangular meshes
        lres = [None, 0.1, 0.01, [0.1, 0.05]]
        lmode = [None, 'rel', 'abs', 'abs']
        lgrid = [None, True, False, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            out = v0.get_sample_mesh(
                res=lres[ii], grid=lgrid[ii], mode=lmode[ii],
            )

        # triangular meshes
        lkeys = ['tri0', 'tri0', 'tri0', 'tri1']
        lres = [None, 0.1, 0.01, [0.1, 0.05]]
        lmode = [None, 'rel', 'abs', 'abs']
        lgrid = [None, True, False, False]
        for ii, k0 in enumerate(lkeys):
            out = self.dobjtri[k0].get_sample_mesh(
                res=lres[ii], grid=lgrid[ii], mode=lmode[ii],
            )

    """
    def test06_sample_bspline(self):
        lres = [None, 0.1, 0.01, [0.1, 0.05]]
        lmode = [None, 'rel', 'abs', 'abs']
        lgrid = [None, True, False, False]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            out = v0.get_sample_bspline(
                res=lres[ii], grid=lgrid[ii], mode=lmode[ii],
            )
    """

    def test07_ev_details_vs_sum(self):

        x = np.linspace(2.2, 2.8, 5)
        y = np.linspace(-0.5, 0.5, 5)
        x = np.tile(x, (y.size, 1))
        y = np.tile(y, (x.shape[1], 1)).T

        # rectangular meshes
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            val = v0.interp2d(
                key=lkey[ii],
                R=x,
                Z=y,
                coefs=None,
                indbs=None,
                indt=None,
                grid=False,
                details=True,
                reshape=True,
                res=None,
                crop=True,
                nan0=ii % 2 == 0,
                imshow=False,
            )
            crop = v0.dobj['bsplines'][lkey[ii]]['crop']
            if crop is False:
                shap = np.prod(v0.dobj['bsplines'][lkey[ii]]['shape'])
            else:
                shap = v0.ddata[crop]['data'].sum()
            assert val.shape == tuple(np.r_[x.shape, shap])

            val_sum = v0.interp2d(
                key=lkey[ii],
                R=x,
                Z=y,
                coefs=None,
                indbs=None,
                indt=None,
                grid=False,
                details=False,
                reshape=True,
                res=None,
                crop=True,
                nan0=ii % 2 == 0,
                imshow=False,
            )
            indok = ~np.isnan(val_sum[0, ...])

            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # Does not work because of knots padding used in func_details
            # Due to scpinterp._bspl.evaluate_spline()...
            if False:   # To be debugged
                assert np.allclose(
                    val_sum[0, indok],
                    np.nansum(val, axis=-1)[indok],
                    equal_nan=True,
                )
            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        # triangular meshes
        lkey = ['tri0-bs0', 'tri1-bs1']
        for ii, (k0, v0) in enumerate(self.dobjtri.items()):
            val = v0.interp2d(
                key=lkey[ii],
                R=x,
                Z=y,
                coefs=None,
                indbs=None,
                indt=None,
                grid=False,
                details=True,
                reshape=None,
                res=None,
                crop=True,
                nan0=ii % 2 == 0,
                imshow=False,
            )
            crop = v0.dobj['bsplines'][lkey[ii]].get('crop', False)
            if crop is False:
                shap = np.prod(v0.dobj['bsplines'][lkey[ii]]['shape'])
            else:
                shap = v0.ddata[crop]['data'].sum()
            assert val.shape == tuple(np.r_[x.shape, shap])

            val_sum = v0.interp2d(
                key=lkey[ii],
                R=x,
                Z=y,
                coefs=None,
                indbs=None,
                indt=None,
                grid=False,
                details=False,
                reshape=None,
                res=None,
                crop=True,
                nan0=ii % 2 == 0,
                imshow=False,
            )
            indok = ~np.isnan(val_sum[0, ...])
            assert np.allclose(
                val_sum[0, indok],
                np.nansum(val, axis=-1)[indok],
                equal_nan=True,
            )

    def test08_plot_mesh(self):

        # rectangular meshes
        lik = [None, ([0, 2], [0, 3]), [2, 3], None]
        lic = [None, ([0, 2], [0, 3]), None, [2, 3]]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            dax = self.dobj[k0].plot_mesh(
                ind_knot=lik[ii],
                ind_cent=lic[ii],
            )
        plt.close('all')

        # triangular meshes
        lik = [None, [0, 2], [2, 3], None]
        lic = [None, [0, 2], None, [2, 3]]
        for ii, (k0, v0) in enumerate(self.dobjtri.items()):
            dax = self.dobjtri[k0].plot_mesh(
                ind_knot=lik[ii],
                ind_cent=lic[ii],
            )
        plt.close('all')

    # TBF for triangular
    def test09_plot_bsplines(self):

        # rectangular meshes
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        lind = [None, ([1, 2], [2, 1]), (1, 1), [1, 2, 10]]
        lknots = [None, True, False, True]
        lcents = [False, False, True, True]
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            dax = self.dobj[k0].plot_bsplines(
                key=lkey[ii],
                ind=lind[ii],
                knots=lknots[ii],
                cents=lcents[ii],
            )
        plt.close('all')

        # triangular meshes
        lkey = ['tri0-bs0', 'tri1-bs1']  # , 'm2-bs2', 'm3-bs3']
        lind = [None, [1, 2], (1, 1), [1, 2, 10]]
        lknots = [None, True, False, True]
        lcents = [False, False, True, True]
        for ii, (k0, v0) in enumerate(self.dobjtri.items()):
            dax = self.dobjtri[k0].plot_bsplines(
                key=lkey[ii],
                ind=lind[ii],
                knots=lknots[ii],
                cents=lcents[ii],
            )
        plt.close('all')

    def test10_plot_profile2d(self):

        # rectangular meshes
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2', 'm3-bs3']
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            key = str(ii)
            kbs = lkey[ii]
            ref = self.dobj[k0].dobj['bsplines'][kbs]['ref']
            shapebs = self.dobj[k0].dobj['bsplines'][kbs]['shape']

            self.dobj[k0].add_data(
                key=key,
                data=np.random.random(shapebs),
                ref=ref,
            )

            dax = self.dobj[k0].plot_profile2d(
                key=key,
            )
        plt.close('all')

        # triangular meshes
        # DEACTIVATED BECAUSE TOO SLOW IN CURRENT VERSION !!!
        if False:
            lkey = ['tri0-bs0', 'tri1-bs1']
            for ii, (k0, v0) in enumerate(self.dobjtri.items()):
                key = str(ii)
                kbs = lkey[ii]
                ref = self.dobjtri[k0].dobj['bsplines'][kbs]['ref']
                shapebs = self.dobjtri[k0].dobj['bsplines'][kbs]['shape']

                self.dobjtri[k0].add_data(
                    key=key,
                    data=np.random.random(shapebs),
                    ref=ref,
                )

                dax = self.dobjtri[k0].plot_profile2d(
                    key=key,
                )
            plt.close('all')

    # TBF for triangular
    def test11_add_bsplines_operator(self):
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2']
        lop = ['D0N1', 'D0N2', 'D1N2', 'D2N2']
        lgeom = ['linear', 'toroidal']
        lcrop = [False, True]

        dfail = {}
        for ii, (k0, v0) in enumerate(self.dobj.items()):
            if ii == 3:
                continue

            for comb in itt.product(lop, lgeom, lcrop):
                deg = self.dobj[k0].dobj['bsplines'][lkey[ii]]['deg']

                # only test exact operators
                if int(comb[0][1]) > deg:
                    # except deg =0 D1N2
                    if deg == 0 and comb[0] == 'D1N2':
                        pass
                    else:
                        continue
                try:
                    self.dobj[k0].add_bsplines_operator(
                        key=lkey[ii],
                        operator=comb[0],
                        geometry=comb[1],
                        crop=comb[2],
                    )
                except Exception as err:
                    dfail[k0] = (
                        f"key {lkey[ii]}, op '{comb[0]}', geom '{comb[1]}': "
                        + str(err)
                    )

        # Raise error if any fail
        if len(dfail) > 0:
            lstr = [f'\t- {k0}: {v0}' for k0, v0 in dfail.items()]
            msg = (
                "The following operators failed:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

    # TBF for triangular
    def test12_compute_plot_geometry_matrix(self):

        # get config and cam
        conf = tf.load_config('WEST-V0')
        cam = tf.geom.utils.create_CamLOS1D(
            pinhole=[3., 1., 0.],
            orientation=[np.pi, 0., 0],
            focal=0.1,
            sensor_nb=50,
            sensor_size=0.15,
            config=conf,
            Diag='SXR',
            Exp='WEST',
            Name='cam1',
        )

        # compute geometry matrices
        for ii, (k0, v0) in enumerate(self.dobj.items()):

            self.dobj[k0].add_geometry_matrix(
                cam=cam,
                res=0.01,
                crop=True,
                store=True,
            )

            dax = self.dobj[k0].plot_geometry_matrix(
                cam=cam, indchan=12, indbf=100,
            )
            plt.close('all')



# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np


# tofu
from . import _generic_check


# #############################################################################
# #############################################################################
#                           Matrix - compute
# #############################################################################


def _compute_check(
    coll=None,
    key=None,
    method=None,
    resMode=None,
    crop=None,
    name=None,
    store=None,
    verb=None,
):

    # key
    lk = list(coll.dobj.get('bsplines', {}).keys())
    if key is None and len(lk) == 1:
        key = lk[0]
    if key not in lk:
        msg = (
            "Arg key must be a valid bspline identifier!\n"
            f"\t- available: {lk}\n"
            f"\t- provided:  {key}"
        )
        raise Exception(msg)

    # method
    method = _generic_check._check_var(
        method, 'method',
        default='los',
        types=str,
        allowed=['los'],
    )

    # resMode
    resMode = _generic_check._check_var(
        resMode, 'resMode',
        default='abs',
        types=str,
        allowed=['abs', 'rel'],
    )

    # crop
    crop = _generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )
    crop = crop and coll.dobj['bsplines'][key]['crop'] is not False

    # name
    if name is None:
        lmat = [
            kk for kk in coll.dobj.get('matrix', {}).keys()
            if kk.startswith('matrix')
        ]
        name = f'matrix{len(lmat)}'
    c0 = (
        isinstance(name, str)
        and name not in coll.dobj.get('matrix', {}).keys()
    )
    if not c0:
        msg = (
            "Arg name must be a str not already taken!\n"
            f"\t- already taken: {coll.dobj.get('matrix', {}).keys()}\n"
            f"\t- provided: {name}"
        )
        raise Exception(msg)

    # store
    store = _generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    # verb
    if verb is None:
        verb = True
    if not isinstance(verb, bool):
        msg = (
            f"Arg verb must be a bool!\n"
            f"\t- provided: {verb}"
        )
        raise Exception(msg)

    return key, method, resMode, crop, name, store, verb


def compute(
    coll=None,
    key=None,
    key_chan=None,
    cam=None,
    res=None,
    resMode=None,
    method=None,
    crop=None,
    name=None,
    store=None,
    verb=None,
):
    """ Compute the geometry matrix using:
            - a Mesh2DRect instance with a key to a bspline set
            - a cam instance with a resolution
    """

    # -----------
    # check input

    key, method, resMode, crop, name, store, verb = _compute_check(
        coll=coll, key=key, method=method, resMode=resMode,
        crop=crop, name=name, store=store, verb=verb,
    )

    # -----------
    # prepare

    nlos = cam.nRays
    shapebs = coll.dobj['bsplines'][key]['shape']

    # -----------
    # compute

    if method == 'los':
        # discretize lines once, then evaluated at points
        pts, reseff, ind = cam.get_sample(
            res=res,
            resMode=resMode,
            DL=None,
            method='sum',
            ind=None,
            pts=True,
            compact=True,
            num_threads=10,
            Test=True,
        )
        lr = np.split(np.hypot(pts[0, :], pts[1, :]), ind)
        lz = np.split(pts[2, :], ind)

        if verb:
            nmax = len(f"Geometry matrix, channel {nlos} / {nlos}")
            nn = 10**(np.log10(nlos)-1)

        # prepare indices
        indbs = coll.select_ind(
            key=key,
            returnas=tuple,
            crop=crop,
        )

        # prepare matrix
        shapemat = tuple(np.r_[nlos, indbs[0].size])
        mat = np.zeros(shapemat, dtype=float)

        for ii in range(nlos):

            # verb
            if verb:
                msg = f"Geom. matrix, chan {ii+1} / {nlos}".ljust(nmax)
                end = '\n' if ii == nlos-1 else '\r'
                print(msg, end=end, flush=True)

            # compute
            mat[ii, :] = np.nansum(
                coll.interp2d(
                    key=key,
                    R=lr[ii],
                    Z=lz[ii],
                    grid=False,
                    indbs=indbs,
                    details=True,
                    reshape=False,
                ),
                axis=0,
            )

        mat = mat * reseff[:, None]
        # scpintg.simps(val, x=None, axis=-1, dx=loc_eff_res[0])

    # -----------
    # return

    if store:

        # add key chan is necessary
        dref = None
        if key_chan is None:
            lrchan = [
                k0 for k0, v0 in coll.dref.items()
                if v0['group'] == 'chan'
                and k0.startswith('chan') and k0[4:].isdecimal()
            ]
            if len(lrchan) == 0:
                chann = 0
            else:
                chann = max([int(k0.replace('chan', '')) for k0 in lrchan]) + 1
            key_chan = f'chan{chann}'

            dref = {
                key_chan: {
                    'data': np.arange(0, nlos),
                    'group': 'chan',
                },
            }

        # add matrix data
        keycropped = f'{key}-cropped' if crop is True else key
        ddata = {
            name: {
                'data': mat,
                'ref': (key_chan, keycropped)
            },
        }

        # add matrix obj
        dobj = {
            'matrix': {
                name: {
                    'bsplines': key,
                    'cam': cam.Id.Name,
                    'data': name,
                    'crop': crop,
                    'shape': mat.shape,
                },
            },
        }

        coll.update(dref=dref, ddata=ddata, dobj=dobj)

    else:
        return mat



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import shutil
import itertools as itt
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu as tf


_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.mesh.test_01_checks'
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_here)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     checking routines
#
#######################################################


class Test01_Inversions():

    @classmethod
    def setup_class(cls):
        pass

    def setup(self):

        # create conf and cam
        conf0 = tf.load_config('WEST-V0')
        cam = tf.geom.utils.create_CamLOS1D(
            pinhole=[3.0, 1., 0.3],
            focal=0.1,
            sensor_size=0.1,
            sensor_nb=30,
            orientation=[-5*np.pi/6, 0, 0],
            config=conf0,
            Name='camH',
            Exp='WEST',
            Diag='SXR',
        )

        # mesh deg 1 and 2
        mesh = tf.data.Mesh2D()
        mesh.add_mesh(
            crop_poly=conf0,
            key='try1',
            res=0.10,
            deg=1,
        )
        mesh.add_bsplines(deg=2)

        # add geometry matrices
        chan = np.arange(0, 30)
        mesh.add_ref(key='chan', data=chan, group='chan')
        mesh.add_geometry_matrix(cam=cam, key='try1-bs1', key_chan='chan')
        mesh.add_geometry_matrix(cam=cam, key='try1-bs2', key_chan='chan')

        # add data
        t0 = np.array([0])
        t1 = np.array([0, 1.])
        data0 = np.exp(-(chan - 15.)**2/10**2)
        data1 = (
            np.exp(-(chan - 15.)**2/10**2)
            + 0.1*np.cos(t1)[:, None]*np.exp(-(chan - 15)**2/2**2)
        )
        # mesh.add_ref(key='t0', data=t0, units='s', group='time')
        mesh.add_ref(key='t1', data=t1, units='s', group='time')
        mesh.add_data(key='data0', data=data0, ref=('chan',))
        mesh.add_data(key='data1', data=data1, ref=('t1', 'chan'))

        self.mesh = mesh

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_run_all_and_plot(self):

        lalgo = tf.data._mesh._inversions_comp._inversions_checks._LALGO
        lstore = [True, False]
        lkdata = ['data0', 'data1']

        # running
        for kmat in ['matrix0', 'matrix1']:

            lop = ['D1N2'] if kmat == 'matrix0' else ['D1N2', 'D2N2']

            for comb in itt.product(lalgo, lkdata, lop, lstore):
                self.mesh.add_inversion(
                    algo=comb[0],
                    key_matrix=kmat,
                    key_data=comb[1],
                    sigma=0.10,
                    operator=comb[2],
                    store=comb[3],
                    conv_crit=1.e-3,
                    kwdargs={'tol': 1.e-4},
                    verb=0,
                )
                ksig = f'{comb[1]}-sigma'
                if ksig in self.mesh.ddata.keys():
                    self.mesh.remove_data(ksig)

        # plotting
        linv = list(self.mesh.dobj['inversions'].keys())[::7]
        for kinv in linv:
            dax = self.mesh.plot_inversion(key=kinv)

        plt.close('all')



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.sparse as scpsp


# specific
from . import _generic_check


_LALGO = [
    'inv_linear_augTikho_sparse',
    'inv_linear_augTikho_dense',
    'inv_linear_augTikho_chol_dense',
    'inv_linear_augTikho_chol_sparse',
    'inv_linear_augTikho_pos_dense',
    'inv_linear_DisPrinc_sparse',
]
_LREGPARAM_ALGO = [
    'augTikho',
    'DisPrinc',
]


# #############################################################################
# #############################################################################
#                           main
# #############################################################################


def _compute_check(
    # input data
    coll=None,
    key_matrix=None,
    key_data=None,
    key_sigma=None,
    data=None,
    sigma=None,
    # choice of algo
    isotropic=None,
    sparse=None,
    positive=None,
    cholesky=None,
    regparam_algo=None,
    algo=None,
    # regularity operator
    solver=None,
    operator=None,
    geometry=None,
    # misc
    conv_crit=None,
    chain=None,
    verb=None,
    store=None,
    # algo and solver-specific options
    kwdargs=None,
    method=None,
    options=None,
):

    # ----
    # keys

    # key_matrix
    lk = list(coll.dobj.get('matrix', {}).keys())
    if key_matrix is None and len(lk):
        key_matrix = lk[0]
    key_matrix = _generic_check._check_var(
        key_matrix, 'key_matrix',
        types=str,
        allowed=lk,
    )
    keybs = coll.dobj['matrix'][key_matrix]['bsplines']
    keym = coll.dobj['bsplines'][keybs]['mesh']
    matrix = coll.ddata[coll.dobj['matrix'][key_matrix]['data']]['data']
    shapemat = matrix.shape
    crop = coll.dobj['matrix'][key_matrix]['crop']

    if np.any(~np.isfinite(matrix)):
        msg = "Geometry matrix should not contain NaNs or infs!"
        raise Exception(msg)

    # key_data
    if key_data is not None or (key_data is None and data is None):
        lk = [
            kk for kk, vv in coll.ddata.items()
            if vv['data'].ndim in [1, 2]
            and vv['data'].shape[-1] == shapemat[0]
        ]
        if key_data is None and len(lk):
            key_data = lk[0]
        key_data = _generic_check._check_var(
            key_data, 'key_data',
            types=str,
            allowed=lk,
        )
        data = coll.ddata[key_data]['data']

    # ------------
    # data, sigma

    # data
    data = _generic_check._check_var(
        data, 'data',
        types=(np.ndarray, list, tuple),
    )
    if not isinstance(data, np.ndarray):
        data = np.asarray(data)
    if data.ndim not in [1, 2] or shapemat[0] not in data.shape:
        msg = (
            "Arg data must have dim in [1, 2]"
            f" and {shapemat[0]} must be in shape\n"
            f"\t- data.shape: {data.shape}"
        )
        raise Exception(msg)
    if data.ndim == 1:
        data = data[None, :]
    if data.shape[1] != shapemat[0]:
        data = data.T
    if np.any(~np.isfinite(data)):
        msg = "Arg data should not contain NaNs or inf!"
        raise Exception(msg)

    # key_sigma
    if key_sigma is not None:
        lk = [
            kk for kk, vv in coll.ddata.items()
            if vv['data'].ndim in [1, 2]
            and vv['data'].shape[-1] == shapemat[0]
        ]
        key_sigma = _generic_check._check_var(
            key_sigma, 'key_sigma',
            types=str,
            allowed=lk,
        )
        sigma = coll.ddata[key_sigma]['data']

    # sigma
    if np.isscalar(sigma):
        sigma = np.full((shapemat[0],), sigma*np.nanmean(np.abs(data)))

    sigma = _generic_check._check_var(
        sigma, 'sigma',
        default=np.full((shapemat[0],), 0.05*np.nanmean(np.abs(data))),
        types=(np.ndarray, list, tuple),
    )

    if not isinstance(sigma, np.ndarray):
        sigma = np.asarray(sigma)
    if sigma.ndim not in [1, 2] or shapemat[0] not in sigma.shape:
        msg = (
            "Arg sigma must have dim in [1, 2]"
            f" and {shapemat[0]} must be in shape\n"
            f"\t- sigma.shape = {sigma.shape}"
        )
        raise Exception(msg)
    if sigma.ndim == 1:
        sigma = sigma[None, :]
    elif sigma.ndim == 2 and data.shape != sigma.shape:
        msg = (
            "Arg sigma must have the same shape as data!\n"
            f"\t- data.shape: {data.shape}\n"
            f"\t- sigma.shape: {sigma.shape}\n"
        )
        raise Exception(msg)
    if sigma.shape[1] != shapemat[0]:
        sigma = sigma.T

    if np.any(~np.isfinite(sigma)):
        msg = "Arg sigma should not contain NaNs or inf!"
        raise Exception(msg)

    # --------------
    # choice of algo

    lc = [
        algo is None,
        all([kk is None for kk in [isotropic, positive, sparse, cholesky]])
    ]

    if not any(lc):
        msg = (
            "Please provide either (xor):\n"
            "\t- algo: directly provide the algo name\n"
            "\t- flags for choosing the algo:\n"
            "\t\t- isotropic: whether to perform isotropic regularization\n"
            "\t\t- sparse: whether to use sparse matrices\n"
            "\t\t- positive: whether to enforce a positivity constraint\n"
            "\t\t- cholesky: whether to use cholesky factorization\n"
        )
        raise Exception(msg)

    if all(lc):
        algo = 'inv_linear_augTikho_sparse'
        lc[0] = False

    if not lc[0] and lc[1]:
        # extract keywrods from algo name
        isotropic = True
        positive = 'pos' in algo
        sparse = 'sparse' in algo
        cholesky = 'chol' in algo

        for aa in _LREGPARAM_ALGO:
            if f'_{aa}_' in algo:
                regparam_algo = aa
                break
        else:
            msg = 'Unreckognized algo for regularization parameter!'
            raise Exception(msg)

    elif lc[0] and not lc[1]:
        # get algo name from keywords

        # isotropic
        isotropic = _generic_check._check_var(
            isotropic, 'isotropic',
            default=True,
            types=bool,
        )
        if isotropic is False:
            msg = "Anisotropic regularization unavailable yet"
            raise NotImplementedError(msg)

        # sparse and matrix and operator
        sparse = _generic_check._check_var(
            sparse, 'sparse',
            default=True,
            types=bool,
        )

        # positive
        positive = _generic_check._check_var(
            positive, 'positive',
            default=False,
            types=bool,
        )

        # cholesky
        cholesky = _generic_check._check_var(
            cholesky, 'cholesky',
            default=False,
            types=bool,
        )
        if positive and cholesky is False:
            msg = "cholesky cannot be used for positive constraint!"
            raise Exception(msg)

        # regparam_algo
        regparam_algo = _generic_check._check_var(
            regparam_algo, 'regparam_algo',
            default='augTikho',
            types=str,
            allowed=_LREGPARAM_ALGO,
        )

        algo = f"inv_linear_{regparam_algo}"
        if cholesky:
            algo += '_chol'
        elif positive:
            algo += '_pos'
        algo += f"_{'sparse' if sparse else 'dense'}"

    # final algo check
    algo = _generic_check._check_var(
        algo, 'algo',
        default=None,
        types=str,
        allowed=_LALGO,
    )

    # -------------------
    # regularity operator

    # get operator
    opmat, operator, geometry, dim, ref, crop = coll.add_bsplines_operator(
        key=keybs,
        operator=operator,
        geometry=geometry,
        returnas=True,
        store=False,
        crop=crop,
    )

    nchan, nbs = matrix.shape
    if isinstance(opmat, tuple):
        assert all([op.shape == (nbs, nbs) for op in opmat])
    elif opmat.ndim == 1:
        msg = "Inversion algorithm requires a quadratic operator!"
        raise Exception(msg)
    else:
        assert opmat.shape == (nbs,) or opmat.shape == (nbs, nbs)
        opmat = (opmat,)

    if not scpsp.issparse(opmat[0]):
        assert all([np.all(np.isfinite(op)) for op in opmat])

    assert data.shape[1] == nchan
    nt = data.shape[0]

    # -------------------
    # consistent sparsity

    # sparse
    if sparse is True:
        if not scpsp.issparse(matrix):
            matrix = scpsp.csc_matrix(matrix)
        if not scpsp.issparse(opmat[0]):
            opmat = [scpsp.csc_matrix(pp) for pp in opmat]
    elif sparse is False:
        if scpsp.issparse(matrix):
            matrix = matrix.toarray()
        if scpsp.issparse(opmat[0]):
            opmat = [scpsp.csc_matrix(pp).toarray() for pp in opmat]

    # -----------------------
    # miscellaneous parameter

    # conv_crit
    conv_crit = _generic_check._check_var(
        conv_crit, 'conv_crit',
        default=1e-4,
        types=float,
    )

    # chain
    chain = _generic_check._check_var(
        chain, 'chain',
        default=True,
        types=bool,
    )

    # verb
    verb = _generic_check._check_var(
        verb, 'verb',
        default=True,
        types=(bool, int),
        allowed=[False, 0, True, 1, 2],
    )
    if verb is False:
        verb = 0
    if verb is True:
        verb = 1

    # store
    store = _generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )
    if key_data is None:
        store = False

    # solver
    solver = _generic_check._check_var(
        solver, 'solver',
        default='spsolve',
        types=str,
        allowed=['spsolve'],
    )

    # ----------------------------------------
    # algo-specific kwdargs and solver options

    # kwdargs, method, options
    kwdargs, method, options = _algo_check(
        algo,
        kwdargs=kwdargs,
        options=options,
        nchan=shapemat[0],
        nbs=shapemat[1],
        conv_crit=conv_crit,
    )

    return (
        key_matrix, key_data, key_sigma, keybs, keym,
        data, sigma, matrix, opmat, operator, geometry,
        isotropic, sparse, positive, cholesky, regparam_algo, algo,
        conv_crit, crop, chain, kwdargs, method, options,
        solver, verb, store,
    )


# #############################################################################
# #############################################################################
#                  ikwdargs / options for each algo
# #############################################################################


def _algo_check(
    algo,
    kwdargs=None,
    method=None,
    options=None,
    nchan=None,
    nbs=None,
    conv_crit=None,
):

    # ------------------------
    # generic kwdargs

    # kwdargs
    if kwdargs is None:
        kwdargs = {}

    # generic kwdargs
    if kwdargs.get('maxiter') is None:
        kwdargs['maxiter'] = 100

    if kwdargs.get('tol') is None:
        kwdargs['tol'] = 1.e-6

    # ------------------------
    # algo specific kwdargs

    # kwdargs specific to aug. tikhonov
    if 'augTikho' in algo:
        a0 = kwdargs.get('a0', 10)
        a1 = kwdargs.get('a1', 2)

        # to have [x]=1
        kwdargs['b0'] = np.math.factorial(a0)**(1 / (a0 + 1))
        kwdargs['b1'] = np.math.factorial(a1)**(1 / (a1 + 1))
        kwdargs['a0'] = a0
        kwdargs['a1'] = a1

        # Exponent for rescaling of a0bis
        # typically in [1/3 ; 1/2], but real limits are 0 < d < 1 (or 2 ?)
        if kwdargs.get('d') is None:
            kwdargs['d'] = 0.95

        if kwdargs.get('conv_reg') is None:
            kwdargs['conv_reg'] = True

        if kwdargs.get('nbs_fixed') is None:
            kwdargs['nbs_fixed'] = True

        if kwdargs['nbs_fixed']:
            kwdargs['a0bis'] = kwdargs['a0'] - 1. + 1200./2.
        else:
            kwdargs['a0bis'] = kwdargs['a0'] - 1. + nbs/2.
        kwdargs['a1bis'] = kwdargs['a1'] - 1. + nchan/2.

    # kwdargs specific to discrepancy principle
    elif 'DisPrinc' in algo:
        if kwdargs.get('chi2n_obj') is None:
            kwdargs['chi2n_obj'] = 1.
        if kwdargs.get('chi2n_tol') is None:
            kwdargs['chi2n_tol'] = 0.05

    # ------------------------
    # low-level solver options

    if 'quad' in algo:

        if options is None:
            options = {}

        if method is None:
            method = 'L-BFGS-B'

        if method == 'L-BFGS-B':
            if options.get('ftol') is None:
                options['ftol'] = conv_crit/100.
            if options.get('disp') is None:
                options['disp'] = False
        else:
            raise NotImplementedError

    return kwdargs, method, options
